#!/usr/bin/perl

# makedepend
# a replacement for the standard makedepend that uses "cc -E" to actually
# preprocess the C source file, then parses its output to generate a
# dependency list.

# John M. Relph <relph@engr.sgi.com>, 12 April 1995
# Copyright 1995 John M. Relph

# see makedepend.html for full usage documentation

# Modified by Rob Lowrie:
#   - Works with source files of the form x.Cx or x.cx, where `x'
#     can be any number of characters.  However, if you name your files
#     things like file.c.c, you deserve what you get.
#   - Calls gcc by default instead of cc
#   - Added `-i' option, which has the same syntax compiler's `-I', but lists 
#     which include directories should actually have dependencies generated.
#     For example `-i. -i/usr/include' says only generate dependencies for
#     include files in directories `/usr/include' and `.' (presumably the
#     directory the Makefile is in).  Note that the `-I' options must still
#     be given in the compiler's options so that it doesn't bomb.
#   - Added -wWidth option (not implemented by Relph), -v, and -h
#   - Unrecognized options print an error and exit.
#   - Prints usage if no source files are given, -h is used, or if an option
#     is not recogonized.
#   - Added errorHandler stuff.

$SIG{INT} = \&errorHandler;  # call errorHandler if you get an interrupt
$SIG{TERM} = \&errorHandler;
$SIG{QUIT} = \&errorHandler;
$SIG{HUP} = \&errorHandler;

# Set this to 1 if you want to see lots of diagnostics

$Debug = 0;

# Set defaults for certain options

#$CC = "gcc -E"; # Compiler to use for parsing
$CC = "cc -E"; # Compiler to use for parsing
$Delimiter =  "# DO NOT DELETE THIS LINE -- make depend depends on it.";
$Prefix = "";
$Suffix = ".o";
$Width = 78;

# parse command line options
&Getopt('Cfopswi','avh');

# help option
if ($opt_h || (! @ARGV)) {
    &printUsage;
    exit;
}

# get -C cc (C compiler and flags to use to compile)
$CC = ($opt_C ? $opt_C : $CC);

# get -f makefile
if ($opt_f) {
    $Makefile = $opt_f;
}
else {
    $Makefile = "makefile";
    if (! (-e $Makefile && -w $Makefile)) {
	$Makefile = "Makefile";
    }

}
if (! (-e $Makefile && -w $Makefile)) {
    die "$0\: cannot open makefile \"$Makefile\"\n";
}

# get -s string
$Delimiter = ($opt_s ? $opt_s : $Delimiter);

# get -p objprefix
$Prefix = ($opt_p ? $opt_p : $Prefix);

# get -o objsuffix
$Suffix = ($opt_o ? $opt_o : $Suffix);

# get -w width
$Width = ($opt_w ? $opt_w : $Width);

# Dump options

if ($opt_v) {
    print "\nmakedepend settings:\n";
    print "   Append (-a): ";
    if ($opt_a) {
	print "true\n";
    }
    else {
	print "false\n";
    }
    print "   Makefile (-f): $Makefile\n";
    print "   Object Suffix (-o): $Suffix\n";
    print "   Object Prefix (-p): $Prefix\n";
    print "   Separation line (-s): $Delimiter\n";
    print "   Width (-w): $Width\n";
    print "   Compiler (-C): $CC\n";
    print "   Defines (-D): $opt_D\n";
    print "   Includes (-I): $opt_I\n";
    print "   Include only dirs (-i):";
    foreach $v (keys %opt_i)
    {
	print " $v";
    }
    print "\n";
    print "   Compiler options (--): $opt_compilerOpts\n\n";
}

rename("$Makefile","$Makefile.bak")
    || die "$0\: cannot make backup \"$Makefile.bak\"\n";

if ($opt_v) {
    print "Original $Makefile moved to $Makefile.bak\n";
}

# copy the existing makefile
open(MAKEIN,"<$Makefile.bak")
    || die "$0\: cannot read backup \"$Makefile.bak\"\n";

open(MAKEOUT,">$Makefile")
    || die "$0\: cannot create makefile \"$Makefile\"\n";

$found = 0;
while (<MAKEIN>) {
    chop;
    last if ($_ eq $Delimiter);
    print MAKEOUT "$_\n";
}

print MAKEOUT "$Delimiter\n";

if ($opt_a) {
    while (<MAKEIN>) {
	chop;
	print MAKEOUT "$_\n";
    }
}

close(MAKEIN);

while (@ARGV) {
    $_ = $ARGV[0];		# get object filename
    if (m/(.*)\.[cC]/) {
	$Obj = "$Prefix$1$Suffix";
    }
    else {
	$Obj = "$Prefix$_$Suffix";
    }

    if ($opt_v) {
	print "$ARGV[0] => $Obj:";
    }

    %Includes = ();

    if ($Debug) {
	print "$CC  $opt_D $opt_I $opt_compilerOpts $ARGV[0]\n";
    }
    open(CPP,"$CC  $opt_D $opt_I $opt_compilerOpts $ARGV[0] | grep '^#' | ");

    while (<CPP>) {
	chop;
# Works with CC:  if (m/^#[0-9]+ "(.+)"/) {
# Modified to work with gcc:
	if (m/^#[ 0-9]+ "(.+)"/) {
	    if ($Debug) {
		print "     Found: ",$1,"\n";
	    }
	    $Includes{$1} = $1;
	}
    }
    close(CPP);

    $Line = "$Obj:";
    $numDepFound = 0;
    $numDepSet = 0;

    foreach $file (sort keys(%Includes)) {
	++$numDepFound;
	($dirSlash = $file) =~ s|[^/]*$||;  # path with trailing /
	($dirNoSlash = $dirSlash) =~ s|/$||;   # path with no trailing /

        if (! $dirSlash) {  # must be in current directory
	    $dirSlash = "./";
	    $dirNoSlash   = ".";
	}

        if (%opt_i == "" ||
	    $opt_i{$dirSlash} || 
            $opt_i{$dirNoSlash}) {

            if ($file ne $ARGV[0]) { # cc -E puts the .c file in, too
		$test = "$Line $file";
		++$numDepSet;

		if (length($test) < $Width) {
		    $Line = $test;
		}
		else {
                    if ($Line ne "$Obj:") {
			print MAKEOUT "$Line\n";
		    }
		    $Line = "$Obj: $file";
		}
	    }
	}
    }

    if ($Line ne "$Obj:") {
	print MAKEOUT "$Line\n";
    }

    if ($opt_v) {
	print " $numDepFound dependencies found";
	if (%opt_i != "") {
	    print ", $numDepSet set";
	}
	print ".\n";
    }

    shift(@ARGV);
}

close(MAKEOUT);

exit;

############################################################################
# taken from Getopt.pl
# modified to save -- args -- separately (as $opt_compilerOpts)
#
# $RCSfile: makedepend.pl,v $$Revision: 1.2 $$Date: 1999/04/09 18:07:49 $
# Process single-character switches with switch clustering.  Pass one argument
# which is a string containing all switches that take an argument, and a second
# argument for those that don't take an argument.  For each
# switch found, sets $opt_x (where x is the switch name) to the value of the
# argument, or 1 if no argument.  Switches which take an argument don't care
# whether there is a space between the switch and the argument.
# Usage:
#   do Getopt('oDI','a');  # -o, -D & -I take arg, while -a is just a switch.

sub Getopt {
    local($argumentative) = $_[0];
    local($switches) = $_[1];
    local($_,$first,$rest);
    local($[) = 0;

    %opt_i = ();

    while (@ARGV && ($_ = $ARGV[0]) =~ /^-(.)(.*)/) {
	($first,$rest) = ($1,$2);

	if ($Debug) {
	    print "first: $first rest: $rest\n";
	}

	if ($first eq "-") {
	    # Compiler options surrounded by --
	    shift(@ARGV);
	    while (@ARGV && ($ARGV[0] ne "--")) {
		# Skip over -o and -c options
		if ($ARGV[0] eq "-o") {
		    shift(@ARGV); # skip outputfile
		}
		elsif ($ARGV[0] ne "-c") {
		    $opt_compilerOpts = "$opt_compilerOpts $ARGV[0]";
		}
		shift(@ARGV);
	    }
	    shift(@ARGV);
	}
	elsif ($first eq "I") {
	    if ($rest ne '') {
		shift(@ARGV);
	    }
	    else {
		shift(@ARGV);
		$rest = shift(@ARGV);
	    }
	    $opt_I = "$opt_I -I$rest";
	}
	elsif ($first eq "i") {
	    if ($rest ne '') {
		shift(@ARGV);
	    }
	    else {
		shift(@ARGV);
		$rest = shift(@ARGV);
	    }
	    $opt_i{$rest} = 1;
	}
	elsif ($first eq "D") {
	    if ($rest ne '') {
		shift(@ARGV);
	    }
	    else {
		shift(@ARGV);
		$rest = shift(@ARGV);
	    }
	    $opt_D = "$opt_D -D$rest";
	}
	elsif (index($argumentative,$first) >= $[) {
	    # This is for options that take an argument
	    if ($rest ne '') {
		shift(@ARGV);
	    }
	    else {
		shift(@ARGV);
		$rest = shift(@ARGV);
	    }
	    eval "\$opt_$first = \$rest;";
	}
	elsif (index($switches,$first) >= $[) {
	    # This is for options that are flags (i.e., no argument)
	    eval "\$opt_$first = 1;";
	    if ($rest ne '') {
		$ARGV[0] = "-$rest";
	    }
	    else {
		shift(@ARGV);
	    }
	}
	else {
	    &printUsage;
	    die "\n*** Option `-$first' not recognized.\n";
	}
    }
}

############################################################################

# printUsage takes no arguments

sub printUsage {
    print "\nUsage: makedepend [options] sourceFile1 [sourceFile2 ...]\n";
    print "\nOptions [default setting]:\n";
    print "   -a            : append results to makefile [false]\n";
    print "   -f Makefile   : makefile name [*tries makefile, then Makefile*]\n";
    print "   -o Suffix     : append Suffix to object file name [$Suffix]\n";
    print "   -p Prefix     : prepend prefix to object file name [$Prefix]\n";
    print "   -s String     : Separation line string [*see documentation*]\n";
    print "   -v            : verbose; prints diagnostics [false]\n";
    print "   -w Width      : maximum width of output line [$Width]\n";
    print "   -C Compiler   : Compiler used to parse source file [$CC]\n";
    print "   -D Defines    : Macro definitions passed to compiler []\n";
    print "   -I Includes   : Include file paths passed to compiler []\n";
    print "   -i Include    : Generate dependencies for only these paths []\n";
    print "   -- Options -- : Compiler options []\n";
}

############################################################################

# error handler.

sub errorHandler {
    my $signame = shift;
    print "\n\n*** makedepend: Received SIG$signame.\n";
    if ($Makefile) {
	print "***   WARNING: $Makefile may be corrupted; check $Makefile.bak\n";
    }
    die "***   Exiting.\n";
}
